/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Written by Denis Lukianov [denis@voxelsoft.com] July 2003 - Royalty Free, No Copyright
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef _CPEFILELOADER_
#define _CPEFILELOADER_

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <windows.h>

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

#define OFFSET_OF_PE_HEADER_IN_STUB			0x3C
#define PE_SECTION_ANSI_NAME_LEN			0x08

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

#pragma pack(1) // This code is very byte-structure sensitive

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

#define SECTION_CHARACTERISTICS_CNT_CODE                   0x00000020
#define SECTION_CHARACTERISTICS_CNT_INITIALIZED_DATA       0x00000040
#define SECTION_CHARACTERISTICS_CNT_UNINITIALIZED_DATA     0x00000080

#define SECTION_CHARACTERISTICS_MEM_16BIT                  0x00020000

#define SECTION_CHARACTERISTICS_ALIGN_1BYTES               0x00100000
#define SECTION_CHARACTERISTICS_ALIGN_2BYTES               0x00200000
#define SECTION_CHARACTERISTICS_ALIGN_4BYTES               0x00300000
#define SECTION_CHARACTERISTICS_ALIGN_8BYTES               0x00400000
#define SECTION_CHARACTERISTICS_ALIGN_16BYTES              0x00500000
#define SECTION_CHARACTERISTICS_ALIGN_32BYTES              0x00600000
#define SECTION_CHARACTERISTICS_ALIGN_64BYTES              0x00700000
#define SECTION_CHARACTERISTICS_ALIGN_128BYTES             0x00800000
#define SECTION_CHARACTERISTICS_ALIGN_256BYTES             0x00900000
#define SECTION_CHARACTERISTICS_ALIGN_512BYTES             0x00A00000
#define SECTION_CHARACTERISTICS_ALIGN_1024BYTES            0x00B00000
#define SECTION_CHARACTERISTICS_ALIGN_2048BYTES            0x00C00000
#define SECTION_CHARACTERISTICS_ALIGN_4096BYTES            0x00D00000
#define SECTION_CHARACTERISTICS_ALIGN_8192BYTES            0x00E00000

#define SECTION_CHARACTERISTICS_LNK_NRELOC_OVFL            0x01000000
#define SECTION_CHARACTERISTICS_MEM_DISCARDABLE            0x02000000
#define SECTION_CHARACTERISTICS_MEM_NOT_CACHED             0x04000000
#define SECTION_CHARACTERISTICS_MEM_NOT_PAGED              0x08000000
#define SECTION_CHARACTERISTICS_MEM_SHARED                 0x10000000
#define SECTION_CHARACTERISTICS_MEM_EXECUTE                0x20000000
#define SECTION_CHARACTERISTICS_MEM_READ                   0x40000000
#define SECTION_CHARACTERISTICS_MEM_WRITE                  0x80000000

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

class CPEFileLoader
{
public:

	struct sPEHeader
	{
		DWORD Signature;
		WORD Machine;
		WORD NumberOfSections;
		DWORD TimeDateStamp;
		DWORD PointerToSymbolTable;
		DWORD NumberOfSymbols;
		WORD SizeOfOptionalHeader;
		WORD Characteristics;
	};

	struct sPEOptionalHeader
	{
		WORD Magic;
		BYTE MajorLinkerVersion;
		BYTE MinorLinkerVersion;
		DWORD SizeOfCode;
		DWORD SizeOfInitializedData;
		DWORD SizeOfUninitializedData;
		DWORD AddressOfEntryPoint;
		DWORD BaseOfCode;
		DWORD BaseOfData;
		DWORD ImageBase;
		DWORD SectionAlignment;
		DWORD FileAlignment;
		WORD MajorOperatingSystemVersion;
		WORD MinorOperatingSystemVersion;
		WORD MajorImageVersion;
		WORD MinorImageVersion;
		WORD MajorSubsystemVersion;
		WORD MinorSubsystemVersion;
		DWORD Reserved0;
		DWORD SizeOfImage;
		DWORD SizeOfHeaders;
		DWORD CheckSum;
		WORD Subsystem;
		WORD DllCharacteristics;
		DWORD SizeOfStackReserve;
		DWORD SizeOfStackCommit;
		DWORD SizeOfHeapReserve;
		DWORD SizeOfHeapCommit;
		DWORD LoaderFlags;
		DWORD NumberOfRvaAndSizes;
		DWORD ExportDirectoryVA;
		DWORD ExportDirectorySize;
		DWORD ImportDirectoryVA;
		DWORD ImportDirectorySize;
		DWORD ResourceDirectoryVA;
		DWORD ResourceDirectorySize;
		DWORD ExceptionDirectoryVA;
		DWORD ExceptionDirectorySize;
		DWORD SecurityDirectoryVA;
		DWORD SecurityDirectorySize;
		DWORD BaseRelocationTableVA;
		DWORD BaseRelocationTableSize;
		DWORD DebugDirectoryVA;
		DWORD DebugDirectorySize;
		DWORD ArchitectureSpecificDataVA;
		DWORD ArchitectureSpecificDataSize;
		DWORD RVAofGPVA;
		DWORD RVAofGPSize;
		DWORD TLSDirectoryVA;
		DWORD TLSDirectorySize;
		DWORD LoadConfigurationDirectoryVA;
		DWORD LoadConfigurationDirectorySize;
		DWORD BoundImportDirectoryinheadersVA;
		DWORD BoundImportDirectoryinheadersSize;
		DWORD ImportAddressTableVA;
		DWORD ImportAddressTableSize;
		DWORD DelayLoadImportDescriptorsVA;
		DWORD DelayLoadImportDescriptorsSize;
		DWORD COMRuntimedescriptorVA;
		DWORD COMRuntimedescriptorSize;
		DWORD Reserved1; 
		DWORD Reserved2;
	};

	struct sPESectionHeader
	{
		char ANSIName[PE_SECTION_ANSI_NAME_LEN];
		DWORD VirtualSize;
		DWORD VirtualAddress;
		DWORD SizeOfRawData;
		DWORD PointerToRawData;
		DWORD PointerToRelocations;
		DWORD PointerToLinenumbers;
		WORD NumberOfRelocations;
		WORD NumberOfLinenumbers;
		DWORD Characteristics;
	};

protected:

	char				*pStub;
	DWORD				StubSize;

	sPEHeader			PEHeader;
	sPEOptionalHeader	PEOptionalHeader;
	
	sPESectionHeader	*pPESections;
	char				**pPESectionsData;

	//sPESectionHeader	*pPEOldFixedSections;

	bool iGetCodeSectionUsingSectionList(unsigned int &iSection, sPESectionHeader *SectionList, unsigned int iSectionsInList);
	bool iGetSectionByVAUsingSectionList(unsigned int iVA, unsigned int &iSection, sPESectionHeader *SectionList, unsigned int iSectionsInList);

public:

	bool Load(char *cFile);
	bool LoadFromResource(HRSRC hResInfo);

	bool CalcOffsets();
	bool Save(char *cFile);
	void Unload();

	sPEHeader *GetHeader();
	sPEOptionalHeader *GetOptionalHeader();
	sPESectionHeader *GetSectionHeader(unsigned int iSection);

	bool GetSection(unsigned int iSection, sPESectionHeader &SectionHeader, char **SectionData);
	//bool SetSection(unsigned int iSection, sPESectionHeader &NewSectionHeader, char *NewSectionData);
	int AppendNewSection(sPESectionHeader &NewSectionHeader, char *NewSectionData);

	unsigned int GetSectionAttributes(unsigned int iSection);
	bool SetSectionAttributes(unsigned int iSection, unsigned int iNewAttributes);

	bool iGetCodeSection(unsigned int &iSection);
	bool iGetSectionByName(char *cName, unsigned int &iSection);
	bool iGetSectionByVA(unsigned int iVA, unsigned int &iSection);

	CPEFileLoader()
	{
		pStub = 0;
		pPESections = 0;
		pPESectionsData = 0;
		//pPEOldFixedSections = 0;
	}
	~CPEFileLoader()
	{
		Unload();
	}
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

#endif // _COMMON_

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
