/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Written by Denis Lukianov [denis@voxelsoft.com] July 2003 - Royalty Free, No Copyright
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "LockEXEFile.h"
#include "Common.h"

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

#define DEBUG_MSGBOX \
	{ \
		char cTemp[2048]; \
		sprintf(cTemp, "%s (%d)", __FILE__, __LINE__); \
		MessageBox(0, cTemp, "Debug notice", MB_OK | MB_ICONSTOP); \
	}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

unsigned int Insert_Scramble(char *cData, unsigned int iDataLen, char *cPassword)
{
	while(iDataLen--)
	{
		cData[iDataLen]++;
	}

	return 1;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool UpdateInsertParameters(char *cPtr,
							unsigned int iInsertVA, unsigned int iInsertCodeSize,
							unsigned int iOriginalCodeVA, unsigned int iOriginalCodeSize,
							unsigned int iOriginalImportAddressTableVA, unsigned int iOriginalImportAddressTableSize,
							unsigned int iNewImportsVA,
							unsigned int iOriginalEntryVA,
							unsigned int iGetCommandLineVA, unsigned int iVirtualProtectVA,
							bool bOriginalEncrypted, char *cPasskey)
{
	// 'FIX' the insert code by correcting all it's data parameters
	int iBytesToSearch = iInsertCodeSize - sizeof(unsigned int);
	unsigned int *iPtr = 0;

	while(iBytesToSearch--)
	{
		iPtr = (unsigned int *)cPtr;

		switch(*iPtr)
		{
		case INSERT_SPECIALCODE_INSERTVA:
			*iPtr = iInsertVA;
			break;
		case INSERT_SPECIALCODE_INSERTSIZE:
			*iPtr = iInsertCodeSize;
			break;
		case INSERT_SPECIALCODE_ORIGINALVA:
			*iPtr = iOriginalCodeVA;
			break;
		case INSERT_SPECIALCODE_ORIGINALSIZE:
			*iPtr = iOriginalCodeSize;
			break;
		case INSERT_SPECIALCODE_ORIGINALIMPORTSADDRESSVA:
			*iPtr = iOriginalImportAddressTableVA;
			break;
		case INSERT_SPECIALCODE_ORIGINALIMPORTSADDRESSSIZE:
			*iPtr = iOriginalImportAddressTableSize;
			break;
		case INSERT_SPECIALCODE_NEWIMPORTSADDRESS:
			*iPtr = iNewImportsVA;
			break;
		case INSERT_SPECIALCODE_ORIGINALENTRYVA:
			*iPtr = iOriginalEntryVA;
			break;
		case INSERT_SPECIALCODE_FUNCTIONVA_GETCOMMANDLINE:
			*iPtr = iGetCommandLineVA;
			break;
		case INSERT_SPECIALCODE_FUNCTIONVA_VIRTUALPROTECT:
			*iPtr = iVirtualProtectVA;
			break;
		case INSERT_SPECIALCODE_WEAKHASH:
			*iPtr = Insert_CalcStringHashWeak(cPasskey);
			break;
		case INSERT_SPECIALCODE_SCRAMBLEHASH:
			srand(LOWORD(GetTickCount()));
			if(bOriginalEncrypted)
				*iPtr = (rand()%0xFFFFFFF0) + 1;
			else
				*iPtr = 0;
			break;
		default:
			break;
		}
		cPtr++;
	}

	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

int LockEXEFile(char *cFileIn, char *cFileOut, char *cPasskey, bool bDoEncrypt, char *cInsert)
{
	CPEFileLoader PEFileInsert;

	if(!PEFileInsert.Load(cInsert))
	{
		MessageBox(0, cInsert, "Missing file:", MB_OK | MB_ICONSTOP);
		return false;
	}

	return LockEXEFile(cFileIn, cFileOut, cPasskey, bDoEncrypt, &PEFileInsert);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

int LockEXEFile(char *cFileIn, char *cFileOut, char *cPasskey, bool bDoEncrypt, unsigned int iRes)
{
	CPEFileLoader PEFileInsert;

	HRSRC ResInfo = FindResource(0, MAKEINTRESOURCE(iRes), "BINARY");

	if(!ResInfo)
		return false;

	if(!PEFileInsert.LoadFromResource(ResInfo))
	{
		MessageBox(0, "", "Missing insert resorce!", MB_OK | MB_ICONSTOP);
		return false;
	}

	return LockEXEFile(cFileIn, cFileOut, cPasskey, bDoEncrypt, &PEFileInsert);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

int LockEXEFile(char *cFileIn, char *cFileOut, char *cPasskey, bool bDoEncrypt, CPEFileLoader *PEFileInsert)
{
	CPEFileLoader PEFileLoader;

	if(!PEFileLoader.Load(cFileIn))
		return false;

	char *cInsertCode;
	char *cOriginalCode;

	unsigned int iInsertCodeSize;
	unsigned int iOriginalCodeVA, iOriginalCodeSize;

	unsigned int iVirtualProtectVA = 0;
	unsigned int iGetCommandLineVA = 0;

	CPEFileLoader::sPEOptionalHeader *OptionalHeader = PEFileLoader.GetOptionalHeader();

	// Load original code to calulate size
	{
		CPEFileLoader::sPESectionHeader Header;
		unsigned int iSection;

		if(!PEFileLoader.iGetCodeSection(iSection))
			return false;

		PEFileLoader.GetSection(iSection, Header, &cOriginalCode);

		iOriginalCodeVA = Header.VirtualAddress;
		iOriginalCodeSize = Header.VirtualSize;
	}
	
	///@ todo check PE header at load time for errors
	/// @todo on UI browse/edit relationship! /// @todo disallow same file in and out
	int iNewImportsSizeEstimate;
	int iNewImportsDirectorySize;
	CImportsTable OriginalImportsTable;
	CImportsTable AdditionalImportsTable;
	// We need to use the GetCommandLine function in the insert, either find it or insert it
	{
		OriginalImportsTable.bExtractImportsTable(&PEFileLoader);
		iNewImportsSizeEstimate = OriginalImportsTable.CalculateImportsSize();
		iNewImportsDirectorySize = OriginalImportsTable.CalculateDirectorySize();
		/// @todo adding 2 functions, same DLL! much overhead to be removed
		if(!OriginalImportsTable.bGetImportFunctionVA("GetCommandLineA", "kernel32.dll", iGetCommandLineVA))
		{
			iNewImportsSizeEstimate += (unsigned int)strlen("GetCommandLineA") + 1 + (unsigned int)strlen("kernel32.dll") + 1 + sizeof(IMAGE_IMPORT_DESCRIPTOR) + sizeof(IMAGE_THUNK_DATA) * 4 + sizeof(unsigned short);
			iNewImportsDirectorySize += sizeof(IMAGE_IMPORT_DESCRIPTOR);
			AdditionalImportsTable.bAddImport("GetCommandLineA", "kernel32.dll", DLL_FUNCTION_HINT_FOR_GETCOMMANDLINEA);
		}
		if(bDoEncrypt)
		{
			if(!OriginalImportsTable.bGetImportFunctionVA("VirtualProtect", "kernel32.dll", iVirtualProtectVA))
			{
				iNewImportsSizeEstimate += (unsigned int)strlen("VirtualProtect") + 1 + /*(unsigned int)strlen("kernel32.dll") + 1 + sizeof(IMAGE_IMPORT_DESCRIPTOR) +*/ sizeof(IMAGE_THUNK_DATA) * 2 + sizeof(unsigned short);
				//iNewImportsDirectorySize += sizeof(IMAGE_IMPORT_DESCRIPTOR); // same dll
				AdditionalImportsTable.bAddImport("VirtualProtect", "kernel32.dll", DLL_FUNCTION_HINT_FOR_VIRTUALPROTECT);
			}
		}
	}
/// @todo nopassword, just encrypt contents option, option to attach to existing section instead of new section
	// 'Encrypt' to make the original EXE invalid without decryption
	{
		if(bDoEncrypt)
			Insert_Scramble(cOriginalCode, iOriginalCodeSize, cPasskey);
	}

	// Load insert code
	{
		CPEFileLoader::sPESectionHeader Header;
		unsigned int iSection;

		if(!PEFileInsert->iGetCodeSection(iSection))
			return false;

		PEFileInsert->GetSection(iSection, Header, &cInsertCode);

		iInsertCodeSize = Header.VirtualSize;
	}

	//unsigned int iOldImportsVA = OptionalHeader->ImportAddressTableVA;

	{
		if(iNewImportsDirectorySize == OptionalHeader->ImportDirectorySize)
			iNewImportsDirectorySize = 0;

		unsigned int iNewSectionSize = iInsertCodeSize + iNewImportsSizeEstimate;
		char *cNewSectionData = new char[iNewSectionSize];
		
		memcpy(cNewSectionData, cInsertCode, iInsertCodeSize);
		
		CPEFileLoader::sPESectionHeader NewSectionHeader;
		strcpy(NewSectionHeader.ANSIName, ".ldata");
		NewSectionHeader.Characteristics = SECTION_CHARACTERISTICS_MEM_EXECUTE | SECTION_CHARACTERISTICS_MEM_READ;
		NewSectionHeader.PointerToLinenumbers = 0;
		NewSectionHeader.NumberOfLinenumbers = 0;
		NewSectionHeader.NumberOfRelocations = 0;
		NewSectionHeader.PointerToRelocations = 0;
		NewSectionHeader.VirtualSize = iNewSectionSize;

		int iNewSection = PEFileLoader.AppendNewSection(NewSectionHeader, cNewSectionData);
		delete[] cNewSectionData;

		PEFileLoader.GetSection(iNewSection, NewSectionHeader, &cNewSectionData);

		// Fix EXE headers
		unsigned int iOldImportAddressTableVA = OptionalHeader->ImportAddressTableVA;
		unsigned int iOldImportAddressTableSize = OptionalHeader->ImportAddressTableSize;
		if(iNewImportsDirectorySize)
		{
			if(iNewImportsDirectorySize)
				bCopyImportsWithAdditions(cNewSectionData + iInsertCodeSize, NewSectionHeader.VirtualAddress + iInsertCodeSize, &OriginalImportsTable, &AdditionalImportsTable);

			OptionalHeader->ImportDirectoryVA = NewSectionHeader.VirtualAddress + iInsertCodeSize; /// @todo thunks in same section! and updare addy VA
			OptionalHeader->ImportDirectorySize = iNewImportsDirectorySize;

			OptionalHeader->ImportAddressTableVA = OptionalHeader->ImportDirectoryVA + iNewImportsDirectorySize;
			//OptionalHeader->ImportAddressTableSize = iNewImportsDirectorySize;///@todo

			AdditionalImportsTable.bGetImportFunctionVA("GetCommandLineA", "kernel32.dll", iGetCommandLineVA);
		}

		// Set original imports to writeable for our app to work on it
		{
			unsigned int iSectionWithImports;
			unsigned int iSectionWithImportsAttributes;
			PEFileLoader.iGetSectionByVA(iOldImportAddressTableVA, iSectionWithImports);
			iSectionWithImportsAttributes = PEFileLoader.GetSectionAttributes(iSectionWithImports);
			iSectionWithImportsAttributes |= SECTION_CHARACTERISTICS_MEM_WRITE;
			PEFileLoader.SetSectionAttributes(iSectionWithImports, iSectionWithImportsAttributes); /// todo temporary solution, use openprocess etc in insert
		}
		// And the code section for decryption
		{
			unsigned int iSectionWithOriginalCode;
			unsigned int iSectionWithOriginalCodeAttributes;
			PEFileLoader.iGetSectionByVA(OptionalHeader->BaseOfCode, iSectionWithOriginalCode);
			iSectionWithOriginalCodeAttributes = PEFileLoader.GetSectionAttributes(iSectionWithOriginalCode);
			iSectionWithOriginalCodeAttributes |= SECTION_CHARACTERISTICS_MEM_WRITE;
			PEFileLoader.SetSectionAttributes(iSectionWithOriginalCode, iSectionWithOriginalCodeAttributes); /// todo temporary solution, use openprocess etc in insert
		}

		// Prevent bound imports from pointing to the optional header [which we modify]
		if(OptionalHeader->BoundImportDirectoryinheadersVA < OptionalHeader->SizeOfHeaders)
		{
			OptionalHeader->BoundImportDirectoryinheadersVA = 0;
			OptionalHeader->BoundImportDirectoryinheadersSize = 0;
		}

		// And write the insert parameters
		//if(!iNewImportsSizeEstimate)///@todo
		UpdateInsertParameters(cNewSectionData,
			NewSectionHeader.VirtualAddress, NewSectionHeader.VirtualSize,
			iOriginalCodeVA, iOriginalCodeSize,
			iOldImportAddressTableVA, iOldImportAddressTableSize,
			OptionalHeader->ImportAddressTableVA,
			OptionalHeader->AddressOfEntryPoint,
			iGetCommandLineVA, iVirtualProtectVA,
			bDoEncrypt, cPasskey);

		// Change entry point
		OptionalHeader->AddressOfEntryPoint = NewSectionHeader.VirtualAddress;
	}

	// Insert our own application into the EXE
	//if(!bInsertCodeIntoCodeSection(iOriginalCodeSize, cCode, iInsertCodeSize, &PEFileLoader))///@todo
	//	return false;

	// We've messed with EXE and possibly moved the imports table, so it needs to be cleaned up
	//if(!bFixImportOffsets(&PEFileLoader, OptionalHeader->ImportAddressTableVA - iOldImportsVA))
	//	return false;

	//if(!bFixResourceOffsets(&PEFileLoader, OptionalHeader->ImportAddressTableVA - iOldImportsVA))
	//	return false;

	// Fiddle entry point to point to us
	if(!PEFileLoader.Save(cFileOut))
		return false;
	
	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
