/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Written by Denis Lukianov [denis@voxelsoft.com] July 2003 - Royalty Free, No Copyright
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "LockEXEFile.h"
#include "CDialog.h"

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <richedit.h>
#include "resource.h"

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifdef MAX_PATH
#undef MAX_PATH
#endif
#define MAX_PATH		8192	// A lot more than windows uses by default

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

int CALLBACK DlgProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg)
	{
	case WM_INITDIALOG:
		SendMessage(hwnd, WM_SETICON, FALSE, (LPARAM)LoadIcon(GetModuleHandle(0), MAKEINTRESOURCE(IDI_ICON1)));
		SendDlgItemMessage(hwnd, IDC_CHECK_DO_ENCRYPT, BM_SETCHECK, BST_CHECKED, 0);
		break;
	case WM_CLOSE:
		EndDialog(hwnd, 0);
		break;
	case WM_COMMAND:
		switch(wParam)
		{
			case IDC_BROWSE_INPUT:
				{
					OPENFILENAME ofn;
					memset(&ofn, 0, sizeof(ofn));

					char cFileTemp[MAX_PATH + 1];
					*cFileTemp = 0;

					ofn.lStructSize = sizeof(ofn);
					ofn.hwndOwner = hwnd;
					ofn.lpstrFilter = "*.exe\0\0";
					ofn.lpstrDefExt = "exe";
					ofn.lpstrFile = cFileTemp;
					ofn.nMaxFile = sizeof(cFileTemp);

					ofn.Flags = OFN_PATHMUSTEXIST | OFN_NOCHANGEDIR | OFN_NOREADONLYRETURN;

					if(GetOpenFileName(&ofn))
					{
						SendDlgItemMessage(hwnd, IDC_EDIT_INPUT, WM_SETTEXT, 0, (LPARAM)(LPSTR)cFileTemp);
					}
				}
				break;
			case IDC_BROWSE_OUTPUT:
				{
					OPENFILENAME ofn;
					memset(&ofn, 0, sizeof(ofn));

					char cFileTemp[MAX_PATH + 1];
					strcpy(cFileTemp, "untitled.exe");

					ofn.lStructSize = sizeof(ofn);
					ofn.hwndOwner = hwnd;
					ofn.lpstrFilter = "*.exe\0\0";
					ofn.lpstrDefExt = "exe";
					ofn.lpstrFile = cFileTemp;
					ofn.nMaxFile = sizeof(cFileTemp);

					ofn.Flags = OFN_NOCHANGEDIR | OFN_NOREADONLYRETURN;

					if(GetSaveFileName(&ofn))
					{
						SendDlgItemMessage(hwnd, IDC_EDIT_OUTPUT, WM_SETTEXT, 0, (LPARAM)(LPSTR)cFileTemp);
					}
				}
				break;
			case IDC_CANCEL:
				{
					SendMessage(hwnd, WM_CLOSE, 0, 0);
				}
				break;
			case IDC_LOCKFILE:
				{
					int iLen = 0;
					char cInput[MAX_PATH];
					char cOutput[MAX_PATH];
					char cPassword[MAX_PATH];
					// Read input
					{
						iLen = (int)SendDlgItemMessage(hwnd, IDC_EDIT_INPUT, WM_GETTEXTLENGTH, 0, 0);
						if(!iLen)
						{
							MessageBox(hwnd, "Need input file", "Input error", MB_OK | MB_ICONEXCLAMATION);
							break;
						}
					}
					SendDlgItemMessage(hwnd, IDC_EDIT_INPUT, WM_GETTEXT, (WPARAM)iLen + 1, (LPARAM)(LPSTR)cInput);
					// Read output
					{
						iLen = (int)SendDlgItemMessage(hwnd, IDC_EDIT_OUTPUT, WM_GETTEXTLENGTH, 0, 0);
						if(!iLen)
						{
							MessageBox(hwnd, "Need output file", "Input error", MB_OK | MB_ICONEXCLAMATION);
							break;
						}
					}
					SendDlgItemMessage(hwnd, IDC_EDIT_OUTPUT, WM_GETTEXT, (WPARAM)iLen + 1, (LPARAM)(LPSTR)cOutput);
					// Do we want to encrypt the contents too?
					bool bDoEncrypt = false;
					if(SendDlgItemMessage(hwnd, IDC_CHECK_DO_ENCRYPT, BM_GETCHECK, 0, 0) == BST_CHECKED)
						bDoEncrypt = true;
					// Attach to last secton or create new section?
					bool bAttachToLastSection = false; /// @todo carry through, enable checkbox
					if(SendDlgItemMessage(hwnd, IDC_CHECK_ATTACH_LASTSECTION, BM_GETCHECK, 0, 0) == BST_CHECKED)
						bAttachToLastSection = true;
					// Read password /// @todo add right click menu to exe files (see shredduh)
					{
						iLen = (int)SendDlgItemMessage(hwnd, IDC_PASSWORD, WM_GETTEXTLENGTH, 0, 0);
						if(!iLen && !bDoEncrypt)
						{
							MessageBox(hwnd, "No point attaching code without passord or encryption\nEither set a password or enable encryption, or both", "Input error", MB_OK | MB_ICONEXCLAMATION);
							break;
						}
					}
					if(iLen)
						SendDlgItemMessage(hwnd, IDC_PASSWORD, WM_GETTEXT, (WPARAM)iLen + 1, (LPARAM)(LPSTR)cPassword);
					else
						strcpy(cPassword, "");
					// Test if password is a alpha-numerical string with no spaces
					if(iLen)
					{
						char *cPtr = (char *)cPassword;
						bool bBadPassword = false;

						do
						{
							char c = *cPtr;
							if(!isalnum(c))
							{
								if(c != '_' || c != '+' || c != '-' || c != '(' || c != ')' || c != '[' || c != ']' || c != '*' || c != '&' || c != '^' || c != '%' || c != '$' || c != '!' || c != '.' || c != ',' || c != '`' || c != '#' || c != '~' || c != '' || c != '@' || c != '\'' || c != '\"' || c != ':' || c != ';' || c != '<' || c != '>' || c != '?' || c != '/' || c != '\\' || c != '|')
								{
									// Bad password
									MessageBox(hwnd, "The password must be an alpha-numeric string with no spaces\nAlso allowed _ + - ( ) [ ] * & ^ % $ ! . , ` # ~  @ \' \" : ; < > ? / \\ |", "Input error", MB_OK | MB_ICONEXCLAMATION);
									bBadPassword = true;
									break;
								}
							}
						}
						while(*++cPtr);

						if(bBadPassword)
							break;
					}
					// DO IT THEN!
					try
					{
						if(LockEXEFile(cInput, cOutput, cPassword, bDoEncrypt, IDR_BINARY1))
						{
							if(MessageBox(hwnd, "The file has been locked successfully\nWould you like to test-run the file?", "File processed", MB_YESNO | MB_ICONINFORMATION) == IDYES)
							{
								ShellExecute(hwnd, "open", cOutput, "", "", SW_SHOW);
								if(iLen)
								{
									if(MessageBox(hwnd, "The locked file has just been run without the password,\ncheck that the original application has not launched.\nWould you like to test with the password?", "Tested without password", MB_YESNO | MB_ICONINFORMATION) == IDYES)
									{
										ShellExecute(hwnd, "open", cOutput, cPassword, "", SW_SHOW);
										MessageBox(hwnd, "The locked file has just been run without the password,\ncheck that the original application was launched successfully.", "Tests complete", MB_OK | MB_ICONINFORMATION);
									}
								}
								else
								{
									MessageBox(hwnd, "The locked file has just been run,\ncheck that the original application was launched successfully.", "Tests complete", MB_OK | MB_ICONINFORMATION);
								}
							}
						}
						else
						{
							MessageBox(hwnd, "Failed to lock file", "ERROR!!!", MB_OK | MB_ICONSTOP);
						}
					}
					catch(...)
					{ /// @todo write protected output no error? also test option on done /// @todo also an option for exe to do a pasword prompt
						MessageBox(hwnd, "Failed to lock file due to an exception", "ERROR!!!", MB_OK | MB_ICONSTOP);
					}
				}
				break;
		}
	default:
		break;
	}

	return false;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	return (int)DialogBoxParam(hInstance, MAKEINTRESOURCE(IDD_DIALOG1), NULL, (DLGPROC)DlgProc, false);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
