/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Written by Denis Lukianov [denis@voxelsoft.com] July 2003 - Royalty Free, No Copyright
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "PEManipulation.h"

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <algorithm>

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool IsDescriptorAhead(sCompleteDescriptorInfo *elem1, sCompleteDescriptorInfo *elem2)
{
	return elem1->ImportDescriptor.FirstThunk < elem2->ImportDescriptor.FirstThunk;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CImportsTable::bExtractImportsTable(CPEFileLoader *PEFileLoader)
{
	CPEFileLoader::sPEOptionalHeader *OptionalHeader = PEFileLoader->GetOptionalHeader();

	CPEFileLoader::sPESectionHeader Header;
	char *cData;
	unsigned int iSection;

	if(!PEFileLoader->iGetSectionByVA(OptionalHeader->ImportDirectoryVA, iSection))
		return false;

	PEFileLoader->GetSection(iSection, Header, &cData);

	int iRelImportAddressTableOffset = OptionalHeader->ImportAddressTableVA - Header.VirtualAddress;
	int iRelImportDirectoryOffset = OptionalHeader->ImportDirectoryVA - Header.VirtualAddress;

	// Read everything into the descriptor list
	IMAGE_IMPORT_DESCRIPTOR *ImportDescriptor = (IMAGE_IMPORT_DESCRIPTOR *)((char *)cData + iRelImportDirectoryOffset);
	while(ImportDescriptor->Characteristics)
	{
		sCompleteDescriptorInfo *CompleteDescriptorInfo = new sCompleteDescriptorInfo;
		CompleteDescriptorInfo->ImportDescriptor = *ImportDescriptor;
		
		char *cNamePointer = (char *)cData + ImportDescriptor->Name - Header.VirtualAddress;
		CompleteDescriptorInfo->cName = new char[strlen(cNamePointer) + 1];
		strcpy(CompleteDescriptorInfo->cName, cNamePointer);
		CompleteDescriptorInfo->bOwnNameString = true;

		IMAGE_THUNK_DATA *OriginalThunkData = (IMAGE_THUNK_DATA *)((char *)cData + ImportDescriptor->OriginalFirstThunk - Header.VirtualAddress);
		IMAGE_THUNK_DATA *ThunkData = (IMAGE_THUNK_DATA *)((char *)cData + ImportDescriptor->FirstThunk - Header.VirtualAddress);

		while(ThunkData->u1.Ordinal && OriginalThunkData->u1.Ordinal)
		{
			sCompleteThunkInfo *CompleteThunkInfo = new sCompleteThunkInfo;
			CompleteThunkInfo->ThunkData = *OriginalThunkData;

			if(IMAGE_SNAP_BY_ORDINAL32(OriginalThunkData->u1.Ordinal))
			{
				//throw; // skip?
			}
			else
			{
				IMAGE_IMPORT_BY_NAME *ThunkByName = (IMAGE_IMPORT_BY_NAME *)((char *)cData + OriginalThunkData->u1.AddressOfData - Header.VirtualAddress);
				CompleteThunkInfo->ThunkByName = *ThunkByName;
				
				CompleteThunkInfo->cName = (char *)ThunkByName->Name;
				CompleteThunkInfo->cName = new char[strlen((char *)ThunkByName->Name) + 1];
				strcpy(CompleteThunkInfo->cName, (char *)ThunkByName->Name);
				CompleteThunkInfo->bOwnNameString = true;
			}

			CompleteDescriptorInfo->ThunkList.push_back(CompleteThunkInfo);
			ThunkData++;
			OriginalThunkData++;
		}

		DescriptorList.push_back(CompleteDescriptorInfo);
		ImportDescriptor++;
	}

	std::sort(DescriptorList.begin(), DescriptorList.end(), IsDescriptorAhead);

	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CImportsTable::bGetImportFunctionVA(char *cFunction, char *cDLL, unsigned int &FunctionVA)
{
	std::vector<sCompleteDescriptorInfo *>::iterator DescriptorIterator = DescriptorList.begin();
	std::vector<sCompleteThunkInfo *>::iterator ThunkInfoIterator;

	while(DescriptorIterator != DescriptorList.end())
	{
		sCompleteDescriptorInfo *Descriptor = *DescriptorIterator;

		if(stricmp(Descriptor->cName, cDLL) == 0)
		{
			ThunkInfoIterator = Descriptor->ThunkList.begin();
			FunctionVA = Descriptor->ImportDescriptor.FirstThunk;
			while(ThunkInfoIterator != Descriptor->ThunkList.end())
			{
				sCompleteThunkInfo *Thunk = *ThunkInfoIterator;
				if(strcmp(Thunk->cName, cFunction) == 0)
				{
					return true;
				}

				FunctionVA += sizeof(IMAGE_THUNK_DATA);
				ThunkInfoIterator++;
			}
		}
		DescriptorIterator++;
	}

	return false;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CImportsTable::bAddImport(char *cFunction, char *cDLL, unsigned short iFunctionHint)
{
	std::vector<sCompleteDescriptorInfo *>::iterator DescriptorIterator;
	std::vector<sCompleteThunkInfo *>::iterator ThunkInfoIterator;
	sCompleteDescriptorInfo *Descriptor = 0;

	DescriptorIterator = DescriptorList.begin();
	while(DescriptorIterator != DescriptorList.end())
	{
		Descriptor = *DescriptorIterator;

		if(stricmp(Descriptor->cName, cDLL) == 0)
		{
			break;
		}
		
		DescriptorIterator++;
	}

	if(DescriptorIterator == DescriptorList.end())
	{
		Descriptor = new sCompleteDescriptorInfo;

		Descriptor->bOwnNameString = true;
		Descriptor->cName = new char[strlen(cDLL) + 1];
		strcpy(Descriptor->cName, cDLL);

		DescriptorList.push_back(Descriptor);
	}

	ThunkInfoIterator = Descriptor->ThunkList.begin();
	sCompleteThunkInfo *Thunk = 0;

	while(ThunkInfoIterator != Descriptor->ThunkList.end())
	{
		Thunk = *ThunkInfoIterator;

		if(strcmp(Thunk->cName, cFunction) == 0)
		{
			return false;
		}

		ThunkInfoIterator++;
	}

	Thunk = new sCompleteThunkInfo;

	Thunk->ThunkByName.Hint = iFunctionHint;
	Thunk->bOwnNameString = true;
	Thunk->cName = new char[strlen(cFunction) + 1];
	strcpy(Thunk->cName, cFunction);

	Descriptor->ThunkList.push_back(Thunk);

	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

unsigned int CImportsTable::CalculateImportsSize()
{
	unsigned int iRet = 0;

	std::vector<sCompleteDescriptorInfo *>::iterator DescriptorIterator;
	std::vector<sCompleteThunkInfo *>::iterator ThunkInfoIterator;

	sCompleteDescriptorInfo *Descriptor = 0;
	sCompleteThunkInfo *Thunk = 0;

	DescriptorIterator = DescriptorList.begin();
	while(DescriptorIterator != DescriptorList.end())
	{
		Descriptor = *DescriptorIterator;

		ThunkInfoIterator = Descriptor->ThunkList.begin();
		
		while(ThunkInfoIterator != Descriptor->ThunkList.end())
		{
			Thunk = *ThunkInfoIterator;

			iRet += sizeof(unsigned short);

			if(!IMAGE_SNAP_BY_ORDINAL32(Thunk->ThunkData.u1.Ordinal))
				if(Thunk->cName)
					iRet += (unsigned int)strlen(Thunk->cName) + 1;

			iRet += sizeof(IMAGE_THUNK_DATA) * 2;

			ThunkInfoIterator++;
		}
		iRet += sizeof(IMAGE_THUNK_DATA);
		iRet += sizeof(IMAGE_THUNK_DATA);

		iRet += (unsigned int)strlen(Descriptor->cName) + 1;
		iRet += sizeof(IMAGE_IMPORT_DESCRIPTOR);

		DescriptorIterator++;
	}
	iRet += sizeof(IMAGE_IMPORT_DESCRIPTOR);

	return iRet;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

unsigned int CImportsTable::CalculateDirectorySize()
{
	return (unsigned int)(DescriptorList.size() + 1) * sizeof(IMAGE_IMPORT_DESCRIPTOR);;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool bCopyImportsWithAdditions(char *cDestination, unsigned int DestinationVA, CImportsTable *Original, CImportsTable *New)
{
	char *cStartOfImports = cDestination;

	std::vector<sCompleteDescriptorInfo *>::iterator DescriptorIterator;
	std::vector<sCompleteThunkInfo *>::iterator ThunkInfoIterator;

	std::vector<sCompleteDescriptorInfo *> MergedImports(Original->DescriptorList);

	DescriptorIterator = New->DescriptorList.begin();
	while(DescriptorIterator != New->DescriptorList.end())
	{
		MergedImports.push_back(*DescriptorIterator);
		DescriptorIterator++;
	}

	sCompleteDescriptorInfo *Descriptor = 0;
	sCompleteThunkInfo *Thunk = 0;
	
	// The new imports do not have valid offsets to thunks, the operation is two-fold:
	// A) first calculate the addresses for the thunks and their names (also the descriptor name!), update in structure
	unsigned int iNewDescriptorsSize = ((unsigned int)MergedImports.size() + 1) * sizeof(IMAGE_IMPORT_DESCRIPTOR);
	char *cThunkOffset = cDestination + iNewDescriptorsSize;
	unsigned int iNewThunksSize = 0;
	DescriptorIterator = MergedImports.begin();
	while(DescriptorIterator != MergedImports.end())
	{
		Descriptor = *DescriptorIterator;
		iNewThunksSize += ((unsigned int)Descriptor->ThunkList.size() + 1) * sizeof(IMAGE_THUNK_DATA);
		DescriptorIterator++;
	}
	char *cOriginalThunkOffset = cThunkOffset + iNewThunksSize;
	char *cNamesOffset = cOriginalThunkOffset + iNewThunksSize;

	// B) secondly write the descriptor, the thunks and all the remaining names
	DescriptorIterator = MergedImports.begin();
	while(DescriptorIterator != MergedImports.end())
	{
		Descriptor = *DescriptorIterator;

		Descriptor->ImportDescriptor.FirstThunk = DestinationVA + (cThunkOffset - cStartOfImports);
		Descriptor->ImportDescriptor.OriginalFirstThunk = DestinationVA + (cOriginalThunkOffset - cStartOfImports);
		Descriptor->ImportDescriptor.ForwarderChain = 0;
		Descriptor->ImportDescriptor.TimeDateStamp = 0;

		ThunkInfoIterator = Descriptor->ThunkList.begin();
		while(ThunkInfoIterator != Descriptor->ThunkList.end())
		{
			Thunk = *ThunkInfoIterator;

			// OriginalThunk
			Thunk->ThunkData.u1.AddressOfData = DestinationVA + (cNamesOffset - cStartOfImports);
			memcpy(cOriginalThunkOffset, &Thunk->ThunkData, sizeof(Thunk->ThunkData));
			cOriginalThunkOffset += (unsigned int)sizeof(Thunk->ThunkData);

			// Thunk's name data
			memcpy(cNamesOffset, &Thunk->ThunkByName.Hint, sizeof(Thunk->ThunkByName.Hint));
			cNamesOffset += (unsigned int)sizeof(Thunk->ThunkByName.Hint);
			memcpy(cNamesOffset, Thunk->cName, strlen(Thunk->cName) + 1);
			cNamesOffset += strlen(Thunk->cName) + 1;

			// Thunk
			memcpy(cThunkOffset, &Thunk->ThunkData, sizeof(Thunk->ThunkData));
			cThunkOffset += (unsigned int)sizeof(Thunk->ThunkData);

			ThunkInfoIterator++;
		}
		IMAGE_THUNK_DATA TerminatorThunk;
		memset(cThunkOffset, 0, sizeof(TerminatorThunk));
		cThunkOffset += (unsigned int)sizeof(TerminatorThunk);
		memset(cOriginalThunkOffset, 0, sizeof(TerminatorThunk));
		cOriginalThunkOffset += (unsigned int)sizeof(TerminatorThunk);

		Descriptor->ImportDescriptor.Name = DestinationVA + (cNamesOffset - cStartOfImports);
		memcpy(cNamesOffset, Descriptor->cName, strlen(Descriptor->cName) + 1);
		cNamesOffset += strlen(Descriptor->cName) + 1;

		memcpy(cDestination, &Descriptor->ImportDescriptor, sizeof(Descriptor->ImportDescriptor));
		cDestination += sizeof(Descriptor->ImportDescriptor);

		DescriptorIterator++;
	}
	IMAGE_IMPORT_DESCRIPTOR TerminatorDescriptor;
	memset(cDestination, 0, sizeof(TerminatorDescriptor));
	cDestination += sizeof(TerminatorDescriptor);

	MergedImports.clear();

	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool GetImportFunctionVA(char *cFunction, char *cDLL, CPEFileLoader *PEFileLoader, unsigned int &ImportFunctionVA)
{
	CPEFileLoader::sPEOptionalHeader *OptionalHeader = PEFileLoader->GetOptionalHeader();

	CPEFileLoader::sPESectionHeader Header;
	char *cData;
	unsigned int iSection;

	if(!PEFileLoader->iGetSectionByVA(OptionalHeader->ImportDirectoryVA, iSection))
		return false;

	PEFileLoader->GetSection(iSection, Header, &cData);

	int iRelImportAddressTableOffset = OptionalHeader->ImportAddressTableVA - Header.VirtualAddress;
	int iRelImportDirectoryOffset = OptionalHeader->ImportDirectoryVA - Header.VirtualAddress;

	// Read everything and compare
	IMAGE_IMPORT_DESCRIPTOR *ImportDescriptor = (IMAGE_IMPORT_DESCRIPTOR *)((char *)cData + iRelImportDirectoryOffset);
	while(ImportDescriptor->Characteristics)
	{
		char *cName = (char *)cData + ImportDescriptor->Name - Header.VirtualAddress;

		if(stricmp(cName, cDLL) != 0)
		{
			ImportDescriptor++;
			continue;
		}

		IMAGE_THUNK_DATA *OriginalThunkData = (IMAGE_THUNK_DATA *)((char *)cData + ImportDescriptor->OriginalFirstThunk - Header.VirtualAddress);
		IMAGE_THUNK_DATA *ThunkData = (IMAGE_THUNK_DATA *)((char *)cData + ImportDescriptor->FirstThunk - Header.VirtualAddress);

		while(OriginalThunkData->u1.Ordinal)
		{
			if(IMAGE_SNAP_BY_ORDINAL32(OriginalThunkData->u1.Ordinal))
			{
				//throw; // skip?
			}
			else
			{
				IMAGE_IMPORT_BY_NAME *ThunkByName = (IMAGE_IMPORT_BY_NAME *)((char *)cData + ThunkData->u1.AddressOfData - Header.VirtualAddress);
				char *cName = (char *)ThunkByName->Name;

				if(strcmp(cName, cFunction) != 0)
				{
					ThunkData++;
					OriginalThunkData++;
					continue;
				}

				ImportFunctionVA = ((unsigned int)((char *)ThunkData - cData)) + Header.VirtualAddress;
				return true;
			}

			ThunkData++;
			OriginalThunkData++;
		}

		ImportDescriptor++;
	}

	return false;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool bAddImport(char *cFunction, char *cDLL, CPEFileLoader *PEFileLoader, unsigned short iHint)
{
	CPEFileLoader::sPEOptionalHeader *OptionalHeader = PEFileLoader->GetOptionalHeader();

	CPEFileLoader::sPESectionHeader Header;
	char *cData;
	unsigned int iSection;

	if(!PEFileLoader->iGetSectionByVA(OptionalHeader->ImportDirectoryVA, iSection))
		return false;

	PEFileLoader->GetSection(iSection, Header, &cData);

	unsigned int iMaxExtraSize = sizeof(IMAGE_IMPORT_DESCRIPTOR)*1 + sizeof(IMAGE_THUNK_DATA)*4 + strlen(cFunction) + strlen(cDLL) + sizeof(unsigned int) + sizeof(unsigned short) + 2;
	unsigned char *cNewData = new unsigned char[Header.VirtualSize + iMaxExtraSize];

	int iRelImportAddressTableOffset = OptionalHeader->ImportAddressTableVA - Header.VirtualAddress;
	int iRelImportDirectoryOffset = OptionalHeader->ImportDirectoryVA - Header.VirtualAddress;

	// Need to somehow preserve section data that does not lie in the import address or directory space
	/*unsigned char *cPreservedResourceData = 0;
	if(OptionalHeader->ResourceDirectoryVA > Header.VirtualAddress && OptionalHeader->ResourceDirectoryVA < Header.VirtualAddress + Header.VirtualSize) /// @todo fix data loss!!!
	{
		cPreservedResourceData = new unsigned char[OptionalHeader->ResourceDirectorySize];
		memcpy(cPreservedResourceData, cData + (OptionalHeader->ResourceDirectoryVA - Header.VirtualAddress), OptionalHeader->ResourceDirectorySize);
	}*/

	// Read everything into the descriptor list
	std::vector<sCompleteDescriptorInfo *> DescriptorList;

	IMAGE_IMPORT_DESCRIPTOR *ImportDescriptor = (IMAGE_IMPORT_DESCRIPTOR *)((char *)cData + iRelImportDirectoryOffset);
	while(ImportDescriptor->Characteristics)
	{
		sCompleteDescriptorInfo *CompleteDescriptorInfo = new sCompleteDescriptorInfo;
		CompleteDescriptorInfo->ImportDescriptor = *ImportDescriptor;
		CompleteDescriptorInfo->cName = (char *)cData + ImportDescriptor->Name - Header.VirtualAddress;

		IMAGE_THUNK_DATA *OriginalThunkData = (IMAGE_THUNK_DATA *)((char *)cData + ImportDescriptor->OriginalFirstThunk - Header.VirtualAddress);
		IMAGE_THUNK_DATA *ThunkData = (IMAGE_THUNK_DATA *)((char *)cData + ImportDescriptor->FirstThunk - Header.VirtualAddress);

		while(ThunkData->u1.Ordinal && OriginalThunkData->u1.Ordinal)
		{
			sCompleteThunkInfo *CompleteThunkInfo = new sCompleteThunkInfo;
			CompleteThunkInfo->ThunkData = *OriginalThunkData;

			if(IMAGE_SNAP_BY_ORDINAL32(OriginalThunkData->u1.Ordinal))
			{
				//throw; // skip?
			}
			else
			{
				IMAGE_IMPORT_BY_NAME *ThunkByName = (IMAGE_IMPORT_BY_NAME *)((char *)cData + OriginalThunkData->u1.AddressOfData - Header.VirtualAddress);
				CompleteThunkInfo->ThunkByName = *ThunkByName;
				CompleteThunkInfo->cName = (char *)ThunkByName->Name;
			}

			CompleteDescriptorInfo->ThunkList.push_back(CompleteThunkInfo);
			ThunkData++;
			OriginalThunkData++;
		}

		DescriptorList.push_back(CompleteDescriptorInfo);
		ImportDescriptor++;
	}

	// Sort DLLs by thunk placement to make sure exe's calls are not disrupted
	std::sort(DescriptorList.begin(), DescriptorList.end(), IsDescriptorAhead);
	
	sCompleteDescriptorInfo *TargetDescriptor;
	{
		TargetDescriptor = new sCompleteDescriptorInfo;

		TargetDescriptor->cName = cDLL;
		TargetDescriptor->ImportDescriptor.TimeDateStamp = 0;
		TargetDescriptor->ImportDescriptor.FirstThunk = 0;
		TargetDescriptor->ImportDescriptor.ForwarderChain = 0;
		TargetDescriptor->ImportDescriptor.Name = 0;
		TargetDescriptor->ImportDescriptor.OriginalFirstThunk = 0;
		TargetDescriptor->ImportDescriptor.Characteristics = 0;

		DescriptorList.push_back(TargetDescriptor);
	}

	sCompleteThunkInfo *TargetThunkInfo;
	{
		TargetThunkInfo = new sCompleteThunkInfo;
		
		TargetThunkInfo->cName = cFunction;
		TargetThunkInfo->ThunkData.u1.AddressOfData = 0;
		TargetThunkInfo->ThunkByName.Hint = iHint;
		TargetThunkInfo->ThunkByName.Name[0] = 0;

		TargetDescriptor->ThunkList.push_back(TargetThunkInfo);
	}

	unsigned int iCurrentDestOffset = 0;
	unsigned int iNewThunkInfoOffset = iCurrentDestOffset;
	unsigned int iNewNumThunks = 0; 

	std::vector<sCompleteThunkInfo *>::iterator ThunkInfoIterator;
	std::vector<sCompleteDescriptorInfo *>::iterator DescriptorIterator = DescriptorList.begin();

	while(DescriptorIterator != DescriptorList.end())
	{
		sCompleteDescriptorInfo *Descriptor = *DescriptorIterator;
		ThunkInfoIterator = Descriptor->ThunkList.begin();
		while(ThunkInfoIterator != Descriptor->ThunkList.end())
		{
			iNewNumThunks++;
			ThunkInfoIterator++;
		}
		iNewNumThunks++; // One for each null termination
		DescriptorIterator++;
	}

	unsigned int iNewDescriptorOffset = iNewThunkInfoOffset + iNewNumThunks * sizeof(IMAGE_THUNK_DATA);
	unsigned int iNewOriginalThunkInfoOffset = iNewDescriptorOffset + (unsigned int)(DescriptorList.size() + 1) * sizeof(IMAGE_IMPORT_DESCRIPTOR);
	unsigned int iNewNameInfoOffset = iNewOriginalThunkInfoOffset + iNewNumThunks * sizeof(IMAGE_THUNK_DATA);

	unsigned int iCurrentThunkInfoOffset = iNewThunkInfoOffset;
	unsigned int iCurrentDescriptorOffset = iNewDescriptorOffset;
	unsigned int iCurrentOriginalThunkInfoOffset = iNewOriginalThunkInfoOffset;
	unsigned int iCurrentNameOffset = iNewNameInfoOffset;

	// Finalise individual offsets as writing to cNewData
	DescriptorIterator = DescriptorList.begin();
	while(DescriptorIterator != DescriptorList.end())
	{
		sCompleteDescriptorInfo *Descriptor = *DescriptorIterator;

		// Set descriptor information
		Descriptor->ImportDescriptor.FirstThunk = Header.VirtualAddress + iCurrentThunkInfoOffset;
		Descriptor->ImportDescriptor.OriginalFirstThunk = Header.VirtualAddress + iCurrentOriginalThunkInfoOffset;
		Descriptor->ImportDescriptor.ForwarderChain = 0;
		Descriptor->ImportDescriptor.TimeDateStamp = 0;

		ThunkInfoIterator = Descriptor->ThunkList.begin();
		while(ThunkInfoIterator != Descriptor->ThunkList.end())
		{
			sCompleteThunkInfo *Thunk = *ThunkInfoIterator;
			
			// Set name address in thunk
			Thunk->ThunkData.u1.AddressOfData = Header.VirtualAddress + iCurrentNameOffset;

			// Write thunkname at that address /// @todo write ordinals too
			memcpy(cNewData + iCurrentNameOffset, &Thunk->ThunkByName.Hint, sizeof(unsigned short));
			memcpy(cNewData + iCurrentNameOffset + sizeof(unsigned short), Thunk->cName, strlen(Thunk->cName) + 1);
			iCurrentNameOffset += (unsigned int)sizeof(unsigned short) + strlen(Thunk->cName) + 1;

			// Write 2 copies of the thunk (and the calculated thunkname offset)
			memcpy(cNewData + iCurrentOriginalThunkInfoOffset, &Thunk->ThunkData, sizeof(IMAGE_THUNK_DATA));
			memcpy(cNewData + iCurrentThunkInfoOffset, &Thunk->ThunkData, sizeof(IMAGE_THUNK_DATA));
			iCurrentThunkInfoOffset += sizeof(IMAGE_THUNK_DATA);
			iCurrentOriginalThunkInfoOffset += sizeof(IMAGE_THUNK_DATA);

			ThunkInfoIterator++;
		}
		// Write 2 terminating thunks
		IMAGE_THUNK_DATA NullThunk;
		NullThunk.u1.Ordinal = 0;
		memcpy(cNewData + iCurrentThunkInfoOffset, &NullThunk, sizeof(IMAGE_THUNK_DATA));
		memcpy(cNewData + iCurrentOriginalThunkInfoOffset, &NullThunk, sizeof(IMAGE_THUNK_DATA));
		iCurrentThunkInfoOffset += sizeof(IMAGE_THUNK_DATA);
		iCurrentOriginalThunkInfoOffset += sizeof(IMAGE_THUNK_DATA);

		// Set name address in descriptor
		Descriptor->ImportDescriptor.Name = Header.VirtualAddress + iCurrentNameOffset;

		// Write descriptor name at that address
		memcpy(cNewData + iCurrentNameOffset, Descriptor->cName, strlen(Descriptor->cName) + 1);
		iCurrentNameOffset += (unsigned int)strlen(Descriptor->cName) + 1;

		// Write descriptor itself
		memcpy(cNewData + iCurrentDescriptorOffset, &Descriptor->ImportDescriptor, sizeof(IMAGE_IMPORT_DESCRIPTOR));
		iCurrentDescriptorOffset += sizeof(IMAGE_IMPORT_DESCRIPTOR);

		// Write descriptor (and the calculated thunk&name offsets), write descriptor name 
		DescriptorIterator++;
	}
	// Write terminating descriptor
	IMAGE_IMPORT_DESCRIPTOR NullDescriptor;
	NullDescriptor.Characteristics = 0;
	NullDescriptor.Name = 0;
	NullDescriptor.ForwarderChain = 0;
	NullDescriptor.OriginalFirstThunk = 0;
	NullDescriptor.TimeDateStamp = 0;
	NullDescriptor.FirstThunk = 0;
	memcpy(cNewData + iCurrentDescriptorOffset, &NullDescriptor, sizeof(IMAGE_IMPORT_DESCRIPTOR));
	iCurrentDescriptorOffset += sizeof(IMAGE_IMPORT_DESCRIPTOR);

	// Restore non-imports data
	/*if(OptionalHeader->ResourceDirectoryVA > Header.VirtualAddress && OptionalHeader->ResourceDirectoryVA < Header.VirtualAddress + Header.VirtualSize)
	{
		memcpy(cNewData + iCurrentNameOffset, cPreservedResourceData, OptionalHeader->ResourceDirectorySize); /// @todo scrap this  - no resources in data section, relocations?
		iCurrentNameOffset += OptionalHeader->ResourceDirectorySize;
		delete[] cPreservedResourceData;
	}*/

	Header.VirtualSize = iCurrentNameOffset; // 5 byte decrease here.. why!?

	//PEFileLoader->SetSection(iSection, Header, (char *)cNewData); /// @todo other data like resources should not perish

	OptionalHeader->ImportAddressTableVA = Header.VirtualAddress;
	OptionalHeader->ImportAddressTableSize = iNewNumThunks * sizeof(IMAGE_THUNK_DATA);
	OptionalHeader->ImportDirectoryVA = Header.VirtualAddress + OptionalHeader->ImportAddressTableSize;
	OptionalHeader->ImportDirectorySize = (unsigned int)(DescriptorList.size() + 1) * sizeof(IMAGE_IMPORT_DESCRIPTOR);
	//OptionalHeader->NumberOfRvaAndSizes ++; /// @todo preservation of debug and resource etc directories because of knock on effects

	// Clean up everything we used
	while(DescriptorList.size())
	{
		delete DescriptorList.back();
		DescriptorList.pop_back();
	}

	delete[] cNewData;

	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool bFixImportOffsets(CPEFileLoader *PEFileLoader, int iOffsetDelta)
{
	// Assumes that PE headers are valid, but the contents of the imports table are off by -iOffsetDelta
	CPEFileLoader::sPEOptionalHeader *OptionalHeader = PEFileLoader->GetOptionalHeader();

	CPEFileLoader::sPESectionHeader Header;
	char *cData;
	unsigned int iSection;

	if(!PEFileLoader->iGetSectionByVA(OptionalHeader->ImportDirectoryVA, iSection))
		return false;

	PEFileLoader->GetSection(iSection, Header, &cData);

	int iRelImportAddressTableOffset = OptionalHeader->ImportAddressTableVA - Header.VirtualAddress; // usually == 0
	int iRelImportDirectoryOffset = OptionalHeader->ImportDirectoryVA - Header.VirtualAddress;

	IMAGE_IMPORT_DESCRIPTOR *ImportDescriptor = (IMAGE_IMPORT_DESCRIPTOR *)((char *)cData + iRelImportDirectoryOffset);
	while(ImportDescriptor->Characteristics)
	{
		ImportDescriptor->Name += iOffsetDelta; // Adjustment

		ImportDescriptor->OriginalFirstThunk += iOffsetDelta; // Adjustment
		IMAGE_THUNK_DATA *OriginalThunkData = (IMAGE_THUNK_DATA *)((char *)cData + ImportDescriptor->OriginalFirstThunk - Header.VirtualAddress);
		ImportDescriptor->FirstThunk += iOffsetDelta; // Adjustment
		IMAGE_THUNK_DATA *ThunkData = (IMAGE_THUNK_DATA *)((char *)cData + ImportDescriptor->FirstThunk - Header.VirtualAddress);

		while(ThunkData->u1.Ordinal && OriginalThunkData->u1.Ordinal)
		{
			if(IMAGE_SNAP_BY_ORDINAL32(OriginalThunkData->u1.Ordinal))
			{
				//throw; // skip
			}
			else
			{
				OriginalThunkData->u1.AddressOfData += iOffsetDelta; // Adjustment
				ThunkData->u1.AddressOfData += iOffsetDelta; // Adjustment
			}

			ThunkData++;
			OriginalThunkData++;
		}

		ImportDescriptor++;
	}

	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

void RecurseFixingResourceOffsets(char *cStartOfResources, IMAGE_RESOURCE_DIRECTORY *Directory, int iRelResourceDirectoryOffset, int iOffsetDelta)
{
	int iEntries = Directory->NumberOfIdEntries + Directory->NumberOfNamedEntries;
	Directory++;
	IMAGE_RESOURCE_DIRECTORY_ENTRY *ResourceDirectoryEntry = (IMAGE_RESOURCE_DIRECTORY_ENTRY *)Directory;
	while(iEntries--)
	{
		if(ResourceDirectoryEntry->DataIsDirectory)
		{
			char *cOffset = cStartOfResources + ResourceDirectoryEntry->OffsetToDirectory;
			IMAGE_RESOURCE_DIRECTORY *ResourceDirectory = (IMAGE_RESOURCE_DIRECTORY *)cOffset;
			RecurseFixingResourceOffsets(cStartOfResources, ResourceDirectory, iRelResourceDirectoryOffset, iOffsetDelta);
		}
		else
		{
			char *cOffset = cStartOfResources + ResourceDirectoryEntry->OffsetToData;
			IMAGE_RESOURCE_DATA_ENTRY *Data = (IMAGE_RESOURCE_DATA_ENTRY *)(cOffset);
			if(Data->OffsetToData)
				Data->OffsetToData += iOffsetDelta;
		}
		ResourceDirectoryEntry++;
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool bFixResourceOffsets(CPEFileLoader *PEFileLoader, int iOffsetDelta)
{
	// Assumes that PE headers are valid, but the contents of the imports table are off by -iOffsetDelta
	CPEFileLoader::sPEOptionalHeader *OptionalHeader = PEFileLoader->GetOptionalHeader();

	CPEFileLoader::sPESectionHeader Header;
	char *cData;
	unsigned int iSection;

	// Nothing to patch?
	if(!OptionalHeader->ResourceDirectorySize)
		return true;

	if(!PEFileLoader->iGetSectionByVA(OptionalHeader->ResourceDirectoryVA, iSection))
		return false;

	PEFileLoader->GetSection(iSection, Header, &cData);

	int iRelResourceDirectoryOffset = OptionalHeader->ResourceDirectoryVA - Header.VirtualAddress;

	IMAGE_RESOURCE_DIRECTORY *ResourceDirectory = (IMAGE_RESOURCE_DIRECTORY *)((char *)cData + iRelResourceDirectoryOffset);

	RecurseFixingResourceOffsets(cData, ResourceDirectory, (int)iRelResourceDirectoryOffset, iOffsetDelta);

	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool bInsertCodeIntoCodeSection(unsigned int iPosition, char *cCode, unsigned int iBytes, CPEFileLoader *PEFileLoader)
{
	CPEFileLoader::sPESectionHeader Header;
	char *cData;

	unsigned int iSection;

	CPEFileLoader::sPEOptionalHeader *OptionalHeader = PEFileLoader->GetOptionalHeader();

	if(!PEFileLoader->iGetCodeSection(iSection))
		return false;

	PEFileLoader->GetSection(iSection, Header, &cData);

	if(Header.VirtualSize < iPosition)
		return false;

	unsigned char *cNewData = new unsigned char[Header.VirtualSize + iBytes];
	unsigned char *cNewDataPtr = cNewData;

	unsigned int iLastBytes = Header.VirtualSize - iPosition;

	if(iPosition)
	{
		memcpy(cNewDataPtr, cData, iPosition);
		cNewDataPtr += iPosition;
	}
	if(iBytes && cCode)
	{
		memcpy(cNewDataPtr, cCode, iBytes);
		cNewDataPtr += iBytes;
	}
	if(iLastBytes)
	{
		memcpy(cNewDataPtr, cData + iPosition, iLastBytes);
		cNewDataPtr += iLastBytes;
	}

	Header.VirtualSize = iPosition + iBytes + iLastBytes;

	//PEFileLoader->SetSection(iSection, Header, (char *)cNewData);

	delete[] cNewData;

	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
