/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Written by Denis Lukianov [denis@voxelsoft.com] July 2003 - Royalty Free, No Copyright
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef _PEMANIPULATION_
#define _PEMANIPULATION_

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "CPEFileLoader.h"

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <vector>

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Structures to manage import descriptors and related information
struct sCompleteThunkInfo
{
	IMAGE_THUNK_DATA		ThunkData;
	IMAGE_IMPORT_BY_NAME	ThunkByName;
	char					*cName;

	bool					bOwnNameString;

	sCompleteThunkInfo()
	{
		cName = 0;
		bOwnNameString = false;
	}

	~sCompleteThunkInfo()
	{
		if(bOwnNameString && cName)
			delete[] cName;
	}
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

struct sCompleteDescriptorInfo
{
	IMAGE_IMPORT_DESCRIPTOR ImportDescriptor;
	char					*cName;

	bool					bOwnNameString;

	std::vector<sCompleteThunkInfo *> ThunkList;

	sCompleteDescriptorInfo()
	{
		cName = 0;
		bOwnNameString = false;
	}

	~sCompleteDescriptorInfo()
	{
		while(ThunkList.size())
		{
			delete ThunkList.back();
			ThunkList.pop_back();
		}

		if(bOwnNameString && cName)
			delete[] cName;
	}
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

class CImportsTable
{
public:

	std::vector<sCompleteDescriptorInfo *> DescriptorList;

	bool bExtractImportsTable(CPEFileLoader *PEFileLoader);
	bool bGetImportFunctionVA(char *cFunction, char *cDLL, unsigned int &FunctionVA);
	bool bAddImport(char *cFunction, char *cDLL, unsigned short iFunctionHint = 0);

	unsigned int CalculateImportsSize();
	unsigned int CalculateDirectorySize();

	CImportsTable()
	{
	}
	~CImportsTable()
	{
		while(DescriptorList.size())
		{
			delete DescriptorList.back();
			DescriptorList.pop_back();
		}
	}
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool IsDescriptorAhead(sCompleteDescriptorInfo *elem1, sCompleteDescriptorInfo *elem2);
bool GetImportFunctionVA(char *cFunction, char *cDLL, CPEFileLoader *PEFileLoader, unsigned int &ImportFunctionVA);
bool bAddImport(char *cFunction, char *cDLL, CPEFileLoader *PEFileLoader, unsigned short iHint = 0);
bool bFixImportOffsets(CPEFileLoader *PEFileLoader, int iOffsetDelta);
bool bFixResourceOffsets(CPEFileLoader *PEFileLoader, int iOffsetDelta);
bool bInsertCodeIntoCodeSection(unsigned int iPosition, char *cCode, unsigned int iBytes, CPEFileLoader *PEFileLoader);
bool bCopyImportsWithAdditions(char *cDestination, unsigned int DestinationVA, CImportsTable *Original, CImportsTable *New);

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

#endif

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
